#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "grafico.h"

/* Inicia sistema grafico Allegro 5.0. 'width' e 'height' é a quantidade
   de unidades (blocos) visuais que devem caber dentro da área de visualização.
   Se parâmetros de entrada são nulos, considera a unidade visual 
   igual a 1 pixel e solicita que usuário escolha resolução.
 */
Window graphinit(int width, int height)
{
  int res_w, res_h, i, modes, max_h, disp_mode;
  Window win = {NULL, NULL, NULL, NULL, {0, 0, 0, 0}};
  ALLEGRO_COLOR default_bg_color = AZUL;
  
  /* Inicializa a Allegro */
  if(!al_init()) {
     fprintf(stderr, "\nNao eh possivel iniciar ambiente grafico Allegro.\n\n");
     exit(2);
  }
  
  al_init_primitives_addon();
  al_init_image_addon();
  al_init_native_dialog_addon();
  al_init_font_addon();
  al_init_ttf_addon();

  al_set_new_display_flags(ALLEGRO_WINDOWED | ALLEGRO_RESIZABLE);   // ALLEGRO_WINDOWED | ALLEGRO_FULLSCREEN_WINDOW
                                                                    // ALLEGRO_FRAMELESS | ALLEGRO_RESIZABLE 
  modes = al_get_num_display_modes();

  if (!width && !height) {
    puts("Escolha uma resolução de janela:");
    for (i=0; i < modes; ++i)
    {
      al_get_display_mode(i, &win.disp_data);
      if (win.disp_data.refresh_rate == 60) // 60 Hz
	printf("(%d): %d %d %d %d\n", i, win.disp_data.width, 
	                                 win.disp_data.height, 
					 win.disp_data.format, 
					 win.disp_data.refresh_rate);
    }
    
    if (i) {  
      scanf("%d", &i);
      getchar();
    }
    else {
      fprintf(stderr, "\nNao eh possivel definir um display para o Allegro.\n\n");
      exit(2);
    }

    al_get_display_mode(i, &win.disp_data);

    win.width_dots = win.disp_data.width; // quantidade de blocos visuais na horizontal
    win.height_dots = win.disp_data.height; // quantidade de blocos visuais na vertical
  }
  else {
    max_h = disp_mode = -1;
    for (i=0; i < modes; ++i)
      {
	al_get_display_mode(i, &win.disp_data);
	if (win.disp_data.refresh_rate == 60) // 60 Hz
	  if (win.disp_data.height > max_h ) {
	    max_h = win.disp_data.height;
	    disp_mode = i;
	  }
      }
    
    al_get_display_mode(disp_mode, &win.disp_data);

    win.width_dots = width; // quantidade de blocos visuais na horizontal
    win.height_dots = height; // quantidade de blocos visuais na vertical
  }

  /* Define tamanho máximo da unidade visual */
  win.dot_width = win.disp_data.width / win.width_dots;
  win.dot_height = win.disp_data.height / win.height_dots;

  /* Define tamanho da área de trabalho */
  res_w = win.width_dots * win.dot_height;
  res_h = win.height_dots * win.dot_height;
  win.display = al_create_display(res_w, res_h);

#ifdef __DEBUG__
    printf("(%d): %d %d %d %d\n", disp_mode, win.disp_data.width, 
	                                     win.disp_data.height, 
	                                     win.disp_data.format, 
	                                     win.disp_data.refresh_rate);
    printf("%d %d %d %d %d %d\n", res_w, res_h, win.dot_width, win.dot_height, win.width_dots, win.height_dots);
#endif // __DEBUG__

  /* Instala o handler de mouse do Allegro */
  al_install_mouse();
  /* Instala o handler de teclado do Allegro */
  al_install_keyboard();
  /* Instala o handler de timer do Allegro */
  win.timer = al_create_timer(1.0 / win.disp_data.refresh_rate);
  
  /* Instala o handler de audio do Allegro */
  if(!al_install_audio()){
    fprintf(stderr, "failed to initialize audio!\n");
    exit(2);
  }
 
  if(!al_init_acodec_addon()){
    fprintf(stderr, "failed to initialize audio codecs!\n");
    exit(2);
  }
 
  if (!al_reserve_samples(1)){
    fprintf(stderr, "failed to reserve samples!\n");
    exit(2);
  }

  /* Define Eventos */
  win.event_queue = al_create_event_queue();
  
  al_register_event_source(win.event_queue, al_get_display_event_source(win.display));
  al_register_event_source(win.event_queue, al_get_mouse_event_source());
  al_register_event_source(win.event_queue, al_get_keyboard_event_source());
  al_register_event_source(win.event_queue, al_get_timer_event_source(win.timer));

  al_clear_to_color(default_bg_color);
  al_flip_display();
  al_start_timer(win.timer);
  al_set_target_bitmap(al_get_backbuffer(win.display));
  
  return (win);
  
}

/* Finaliza sistema grafico Allegro 5.0 */
void graphdeinit(Window win)
{
  al_destroy_sample(win.som);
  al_destroy_timer(win.timer);
  al_destroy_event_queue(win.event_queue);
  al_destroy_display(win.display);
}


/* Cria  um cursor (bloco retangular) em ambiente ALLEGRO  */
ALLEGRO_BITMAP *blocoBitmap(int width, int height, ALLEGRO_COLOR color, Window win)
{
  ALLEGRO_BITMAP *bmp, *old_target;

  old_target = al_get_target_bitmap();
  
  bmp = al_create_bitmap(width, height);
  al_set_target_bitmap(bmp);
  al_clear_to_color(color);

  // al_set_target_bitmap(al_get_backbuffer(win.display));
  al_set_target_bitmap(old_target);
  
  return bmp;
}


/* Cria  região em  display para  exibir imagem PNG  ou JPG  em ambiente
   ALLEGRO a partir de arquivo
 */
ALLEGRO_BITMAP *imagemArq(char *nomeArqImg, int width, int height, Window win)
{
  ALLEGRO_BITMAP *bmp, *img, *old_target;
  int iw, ih;
  
  old_target = al_get_target_bitmap();
  
  bmp = blocoBitmap(width, height, NEVE, win);
  al_set_target_bitmap(bmp);
  
  if (nomeArqImg) {
    img = al_load_bitmap(nomeArqImg);
    iw = al_get_bitmap_width(img);
    ih = al_get_bitmap_height(img);
    // al_draw_bitmap(img,0,0,0);
    al_draw_scaled_bitmap(img,0,0,iw,ih,0,0,width,height,0);
  }
  
  // al_set_target_bitmap(al_get_backbuffer(win.display));
  al_set_target_bitmap(old_target);
  
  return bmp;
}


/* Cria bitmap para caber 'texto'.
   Bitmap terá dimensão 'width x height' ou igual à dimensão da fonte, 
   dependendo de quem for maior.
   Retorna um TEXMAP contendo o bitmap pronto para ser preenchido com 'texto'.
*/
TEXTMAP *blocoTexto(char *texto, ALLEGRO_FONT *fonte, ALLEGRO_COLOR cor_bloco, ALLEGRO_COLOR cor_texto, int width, int height, Window win)
{
  TEXTMAP *img = (TEXTMAP *) calloc(1,sizeof(TEXTMAP));
  
  img->txt = strdup(texto);
  img->font = fonte;
  
  img->bloco.w = al_get_text_width(img->font, img->txt);
  img->bloco.w = (width > img->bloco.w) ? width : img->bloco.w;

  img->bloco.h = al_get_font_line_height(img->font);
  img->bloco.h = (height > img->bloco.h) ? height : img->bloco.h;
  
  img->bloco.bmp = blocoBitmap(img->bloco.w, img->bloco.h, cor_bloco, win);
 
  return img;
}


/* Cria um bitmap com 'texto', centralizado no bitmap.
   Usa função 'blocoTexto()' para criar bitmap e depois
   desenha 'texto' dentro.
   Retorna TEXTMAP contendo o bitmap com o texto dentro.
*/
TEXTMAP *centerText(char *texto, ALLEGRO_FONT *fonte, ALLEGRO_COLOR cor_bloco, ALLEGRO_COLOR cor_texto, int width, int height, Window win)
{
  TEXTMAP *img;
  ALLEGRO_BITMAP *old_target;

  old_target = al_get_target_bitmap();

  img = blocoTexto(texto, fonte, cor_bloco, cor_texto, width, height, win);
  
  al_set_target_bitmap(img->bloco.bmp);
  al_draw_text(img->font, cor_texto, img->bloco.w/2.0, (img->bloco.h - al_get_font_ascent(img->font))/2.0, ALLEGRO_ALIGN_CENTRE, img->txt);
  
  // al_set_target_bitmap(al_get_backbuffer(win.display));
  al_set_target_bitmap(old_target);

  return img;

}


/* Cria um bitmap com 'texto', alinhado à esquerda no bitmap.
   Usa função 'blocoTexto()' para criar bitmap e depois
   desenha 'texto' dentro.
   Retorna TEXTMAP contendo o bitmap com o texto dentro.
*/
TEXTMAP *leftText(char *texto, ALLEGRO_FONT *fonte, ALLEGRO_COLOR cor_bloco, ALLEGRO_COLOR cor_texto, int width, int height, Window win)
{
  TEXTMAP *img;
  ALLEGRO_BITMAP *old_target;

  old_target = al_get_target_bitmap();

  img = blocoTexto(texto, fonte, cor_bloco, cor_texto, width, height, win);
  
  al_set_target_bitmap(img->bloco.bmp);
  al_draw_text(img->font, cor_texto, 0.0, al_get_font_ascent(img->font)/2.0, ALLEGRO_ALIGN_LEFT, img->txt);
  
  // al_set_target_bitmap(al_get_backbuffer(win.display));
  al_set_target_bitmap(old_target);

  return img;

}

/* Cria região em display para exibir imagem em ambiente ALLEGRO a partir 
   uma imagem PPM. Esta consiste de um vetor linear de valores do tipo 
   'Pixel', representando uma matriz width x height.
   Se arg. 'imgPPM' é NULL, apaga imagem atual e BITMAP fica na cor NEVE.
 */
ALLEGRO_BITMAP *imagemPPM(Pixel *imgPPM, int width, int height, Window win)
{
  ALLEGRO_BITMAP *bmp, *old_target;
  int i,x,y;

  old_target = al_get_target_bitmap();
  
  bmp = blocoBitmap(width, height, NEVE, win);
  al_set_target_bitmap(bmp);
  
  if(imgPPM) {
    for (i=x=y=0; i < width*height; ++i) {
      y = i / width;
      x = i % width;
      al_put_pixel(x, y, BPIXEL(imgPPM[i]));
    }
  }


  // al_set_target_bitmap(al_get_backbuffer(win.display));
  al_set_target_bitmap(old_target);

  return bmp;
}


/* Muda a imagem em um BITMAP para imagem em formato PPM. 
   Se arg. 'imgPPM' é NULL, apaga imagem atual e BITMAP fica na cor NEVE.
 */
void mudaImgPPM(BITMAP imagem, Pixel *imgPPM, int width, int height, Window win)
{
  ALLEGRO_BITMAP *old_target;
  int i,x,y;

  old_target = al_get_target_bitmap();
  
  al_set_target_bitmap(imagem.bmp);

  if(imgPPM) {
    for (i=x=y=0; i < width*height; ++i) {
      y = i / width;
      x = i % width;
      al_put_pixel(x, y, BPIXEL(imgPPM[i]));
    }
  }
  else
    al_clear_to_color(NEVE);

  // al_set_target_bitmap(al_get_backbuffer(win.display));
  al_set_target_bitmap(old_target);

  al_draw_bitmap(imagem.bmp, imagem.x, imagem.y, 0);
}


/* Retorna 1 se as coordenadas x,y estão dentro da imagem, 0 caso contrário. */
bool is_inside_bitmap (BITMAP img, int x, int y)
{
  return (x >= img.x && x < img.x+img.w &&
	  y >= img.y && y < img.y+img.h );
}

/* Altera o PERÍODO do timer de Window, multiplicando o período 
   atual pelo fator.
   Nota: Ao criar Window, o período inicial é definido como sendo 
         '1.0 / win.disp_data.refresh_rate'
*/
void setTimerSpeed(Window win, double fator)
{
  al_set_timer_speed(win.timer, al_get_timer_speed(win.timer) * fator);

}

